# composer_gui.py
#
# Copyright 2006-2007 Helsinki Institute for Information Technology
# and the authors.
#
# Authors: Ville Mäntysaari <ville.mantysaari@hiit.fi>
#          Pekka Kanerva <kanerva@hiit.fi>
#          Tero Hasu <tero.hasu@hut.fi>
#

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import e32
import appuifw

from s4all.composer.util import Str, Config, Util
from s4all.composer.composer_action_store import ActionStore
from s4all.composer.composer_workflow import Workflow


############################################################
#
#                     GUI
#
############################################################

class GUI:
    """
    GUI contains the main class for run the application
    """

    def __init__(self):
        self.lock = e32.Ao_lock()
        self.main_title = u"Composer"
        self.old_title = appuifw.app.title
        appuifw.app.title = self.main_title

        self.wflow = None
        
        self.main_view = appuifw.Listbox([self.__wflow_name()],
                                         lambda: None)
        appuifw.app.body = self.main_view

        self.main_menu = [(u"Create workflow",
                           self.new_wflow_handler),
                          (u"Load workflow",
                           self.load_wflow_handler),
                          (u'Remove local actions',
                           self.remove_local_actions_handler)]

        appuifw.app.menu = self.main_menu
        appuifw.app.exit_key_handler = self.abort

        self.a_store = ActionStore()

    def remove_local_actions_handler(self):
        self.a_store.remove_local_actions()

    def close_wflow_handler(self):
        self.wflow = None
        self.main_view.set_list([self.__wflow_name()])
        appuifw.app.title = self.main_title
        appuifw.app.menu = self.main_menu

    def load_wflow_handler(self):
        actlist = [ act for act in self.a_store.get_action_list()
                    if act["data"].get("type", None) == "wflow" ]
        if len(actlist) == 0:
            appuifw.note(u"No workflows to load", "error")
            return
        namelist = [ unicode(wf["data"]["name"]) for wf in actlist ]
        # Note that passing an empty list appears to cause KERN-EXEC 3.
        if namelist == []:
            appuifw.note(u"No saved workflows", "error")
            return
        index = appuifw.popup_menu(namelist, u'Select workflow')
        if index is None:
            return

        new_wflow = Workflow(namelist[index], self, self.a_store)

        try:
            new_wflow.load(actlist[index]["file"])
        except:
            # Assuming "load" reports errors.
            return

        self.wflow = new_wflow

        appuifw.app.menu = self.wflow.wflow_menu
        self.main_view.set_list(self.wflow.get_action_names())
        appuifw.app.title = self.__wflow_name()

    def new_wflow_handler(self):
        new_name = appuifw.query(u"Workflow name",
                                 "text",
                                 u"workflow")
        if new_name is None:
            return

        basename = Util.wflow_basename_from_name(new_name)
        if not Util.wflow_exists_check(basename):
            return

        self.wflow = Workflow(new_name, self, self.a_store)

        appuifw.app.menu = self.wflow.wflow_menu
        appuifw.app.title = self.__wflow_name()
        self.main_view.set_list( self.wflow.get_action_names() )

    def abort(self):
        self.lock.signal()

    def loop(self):
        self.lock.wait()

        if self.wflow is not None:
            self.wflow.close_wflow_handler()

    def close(self):
        appuifw.app.menu = []
        appuifw.app.body = None
        appuifw.app.exit_key_handler = None
        appuifw.app.title = self.old_title

    def __wflow_name(self):
        if self.wflow is None:
            return Str.no_wflow
        return self.wflow.name


def main():
    gui = GUI()
    try:
        gui.loop()
    finally:
        gui.close()


if __name__ == "__main__":
    main()
