# composer_action_store.py
#
# Copyright 2006-2007 Helsinki Institute for Information Technology
# and the authors.
#
# Authors: Ville Mäntysaari <ville.mantysaari@hiit.fi>
#          Pekka Kanerva <kanerva@hiit.fi>
#          Tero Hasu <tero.hasu@hut.fi>
#

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import e32
import appuifw

import os, re, urllib

from s4all.composer.util import Str, Config, Util


############################################################
#
#                     ActionStore
#
############################################################

class ActionStore:
    
    """
    A storage/retrieval class to retrieve/store actions from/to various places.

    """

    def __init__(self):

        # Our model contains entries of the form
        # {"file": basename,
        #  "module": modref (if loaded),
        #  "data": action data}.
        # Indexing is by basename.

        self.actmap = {}
        
        self.refresh_local_actions()

    def interactive_find_action(self):
        """
        Returns the chosen action entry, or None if there were no
        actions to choose from, or if none was selected.
        """
        if len(self.actmap) == 0:
            self.__error_no_actions()
            return None

        criteria = appuifw.query(u'Enter search string', 'text', u'')
        if criteria is None:
            return None

        results = []
        criteria = criteria.lower()
        rexp = re.compile("\\b" + re.escape(criteria), re.IGNORECASE)
        for act in self.actmap.values():
            data = act["data"]
            if data["name"].lower().startswith(criteria):
                results.append(act)
            else:
                found = False
                if "tags" in data:
                    for tag in data["tags"]:
                        if tag.lower().startswith(criteria):
                            found = True
                            break
                if (not found) and ("doc" in data):
                    if rexp.search(data["doc"]):
                        found = True
                if found:
                    results.append(act)

        if results == []:
            appuifw.note(u'No matching action found', 'error')
            return None

        self.__sort_actlist(results)
        names = [ unicode(act["data"]["name"]) for act in results ]
        index = appuifw.popup_menu(names, u'Select action')
        if index is None:
            return None
        
        return results[index]

    def get_action_list(self):
        list = self.actmap.values()
        self.__sort_actlist(list)
        return list

    def get_module(self, acttype, action):
        """
        Gets something that looks like a module for the specified
        action. This involves loading or constructing the module if
        that has not already been done.

        Note that "acttype" is begin passed as this function may be
        used during bootstrapping, and at that time we cannot
        necessarily assume that "data" is already available.
        """

        if "module" not in action:
            if acttype == "wsdl":
                from soappy import WSDL
                pathname = Config.wsdl_dir + action["file"]
                server = WSDL.Proxy(pathname)
                action["module"] = WSDLWrapper(server, pathname)
            elif acttype == "python" or acttype == "control":
                self._import_action(action)
            else:
                assert False
            
        return action["module"]
        
    def get_action_names(self):
        """
        Returns the names of actions in the store (for GUI display).
        """
        if len(self.actmap) == 0:
            return [ Str.no_actions ]

        return [ a["data"]["name"] for a in self.actmap.values() ]

    def get_action_basenames(self):
        """
        Returns the basenames of the action files.
        """
        return self.actmap.keys()

    def get_action_by_basename(self, basename):
        """
        The basename can be either Unicode or non-Unicode, and should
        match the relevant record regardless.
        """
        return self.actmap[basename]

    def store_action(self, basename, actdata):
        entry = {"file": basename,
                 "data": actdata}
        self.actmap[basename] = entry

    def refresh_local_actions(self):
        """
        Updates the internal list of actions. This involves
        a rescan of the directories in which the actions are stored.
        """

        scanmap = {}
        for file in Util.action_files():
            scanmap[file] = {"type": "python",
                             "file": file}
        for file in Util.wsdl_files():
            scanmap[file] = {"type": "wsdl",
                             "file": file}

        oldmap = {}
        for file, act in self.actmap.iteritems():
            if file in scanmap:
                oldmap[file] = act

        new_wflows = []
        for file, act in scanmap.iteritems():
            if file not in oldmap:
                oldmap[file] = act

                acttype = act["type"]
                if acttype == "wsdl":
                    act["data"] = self._get_wsdl_data(act)
                elif acttype == "python":
                    act["data"] = self._get_action_data(
                        Config.actions_dir + file)
                    if act["data"]["type"] == "wflow":
                        new_wflows.append(act)
                    elif act["data"]["type"] == "control":
                        act["type"] = "control"
                    elif act["data"]["type"] == "complex":
                        act["type"] = "complex"
                else:
                    assert False

                actdata = act["data"]
                if "type" not in actdata:
                    actdata["type"] = acttype
                if "name" not in actdata:
                    actdata["name"] = unicode(file)

        """
        # Set "union tags" for newly loaded workflows now that
        # hopefully all of their dependencies have been loaded, too.
        # XXX this does not quite do the job if the newly loaded workflows contain newly loaded workflows
        for act in new_wflows:
            try:
                actdata = act["data"]
                tags_dict = {}
                for a in actdata["actions"]:
                    sub_act = oldmap[a["file"]]
                    subdata = sub_act["data"]
                    if "tags" in subdata:
                        for t in subdata["tags"]:
                            tags_dict[t] = 1

                actdata["tags"] = tags_dict.keys()
            except KeyError:
                # Not all of the dependencies were loaded, I guess.
                pass
        """
        self.actmap = oldmap

    def remove_local_actions(self):
        if len(self.actmap) == 0:
            self.__error_no_actions()
            return

        actlist = self.get_action_list()
        namelist = [ a["data"]["name"] for a in actlist ]

        indexes = appuifw.multi_selection_list(namelist)
        if len(indexes) == 0:
            return

        really = appuifw.query(u'Remove selected actions?', 'query')
        if really:
            for action in [ actlist[i] for i in indexes ]:
                if ((action["data"]["type"] == "python") or
                    (action["data"]["type"] == "wflow")):
                    os.remove(Config.actions_dir + action["file"])
                elif action["data"]["type"] == "wsdl":
                    file = Config.wsdl_dir + action["file"]
                    os.remove(file)
                    try:
                        os.remove(Util.wsdl_cache_file(file))
                    except OSError:
                        pass
                else:
                    assert False
            self.refresh_local_actions()
        else:
            appuifw.note(u'Removing actions cancelled', 'info')

    def _search_actions_handler(self):

	"""
          Search new actions from a UDDI server.
        """

        from uddi4py.client import UDDIProxy
        from uddi4py.requestDOM import findService
        from uddi4py.requestDOM import getServiceDetail
        from uddi4py.requestDOM import getTModelDetail
        from uddi4py.baseDOM import UDDIError
        
        criteria = appuifw.query(u'Enter search criteria', 'text', u'')

        if criteria is None or criteria == '':
            return

        iURL = "http://uddi.xmethods.net/inquire"
        pURL = "https://uddi.xmethods.net/publish"

        try:
            uddiCon = UDDIProxy(iURL, pURL)

            request = findService()
            request.setNameStrings(criteria)

            responseDOM = uddiCon.send(request)
            #response = serviceList(responseDOM)

            services = responseDOM.getElementsByTagName("serviceInfo")

            list = []
            serviceName = ''

            for serv in services:
                name = serv.getElementsByTagName("name")
                list.append(unicode(name[0].firstChild.data))

            index = appuifw.selection_list(list)

            selected = services[index]
            serviceName = list[index]
            
            sKey = selected.getAttribute("serviceKey")
            
            request = getServiceDetail()
            request.setServiceKeyStrings([sKey])
            
            responseDOM = uddiCon.send(request)
            
            tModelDet = responseDOM.getElementsByTagName("tModelInstanceDetails")
            tModelKey = tModelDet[0].firstChild.getAttribute("tModelKey")
            
            request = getTModelDetail()
            request.setTModelKeyStrings([tModelKey])
            
            responseDOM = uddiCon.send(request)
            
            overview = responseDOM.getElementsByTagName("overviewURL")
            wsdlURL = overview[0].firstChild.data
            
            wsdlURL = wsdlURL.split('#')
            
            urllib.urlretrieve(wsdlURL[0], Config.wsdl_dir + serviceName.replace(' ', '') + '.wsdl')


        except(UDDIError),e:
            return u'Error'
    

        self.refresh_local_actions()
        appuifw.note(u'Download finished', 'info')
                               
    def __sort_actlist(self, actlist):
        """
        Note that this method sorts the given list in place, and
        returns nothing.
        """
        actlist.sort(lambda x, y: cmp(x["data"]["name"],
                                      y["data"]["name"]))

    def get_action_names(self):
        """
        Returns the names of actions in the store (for GUI display).
        """
        
        if len(self.actmap) == 0:
            return [ Str.no_actions ]

        return [ a["data"]["name"] for a in self.actmap.values() ]

    def get_action_basenames(self):
        """
        Returns the basenames of the action files.
        """

        return self.actmap.keys()

    def get_action_by_basename(self, basename):
        """
        The basename can be either Unicode or non-Unicode, and should
        match the relevant record regardless.
        """
        return self.actmap[basename]

    def store_action(self, basename, actdata):
        entry = {"file": basename,
                 "data": actdata}
        self.actmap[basename] = entry

    def refresh_local_actions(self):
        """
        Updates the internal list of actions. This involves
        a rescan of the directories in which the actions are stored.
        """

        scanmap = {}
        for file in Util.action_files():
            scanmap[file] = {"type": "python",
                             "file": file}
        for file in Util.wsdl_files():
            scanmap[file] = {"type": "wsdl",
                             "file": file}

        oldmap = {}
        for file, act in self.actmap.iteritems():
            if file in scanmap:
                oldmap[file] = act

        new_wflows = []
        for file, act in scanmap.iteritems():
            if file not in oldmap:
                oldmap[file] = act

                acttype = act["type"]
                if acttype == "wsdl":
                    act["data"] = self._get_wsdl_data(act)
                elif acttype == "python":
                    act["data"] = self._get_action_data(
                        Config.actions_dir + file)
                    if act["data"]["type"] == "wflow":
                        new_wflows.append(act)
                    elif act["data"]["type"] == "control":
                        act["type"] = "control"
                else:
                    assert False

                actdata = act["data"]
                if "type" not in actdata:
                    actdata["type"] = acttype
                if "name" not in actdata:
                    actdata["name"] = unicode(file)

        """
        # Set "union tags" for newly loaded workflows now that
        # hopefully all of their dependencies have been loaded, too.
        # XXX this does not quite do the job if the newly loaded workflows contain newly loaded workflows
        for act in new_wflows:
            try:
                actdata = act["data"]
                tags_dict = {}
                for a in actdata["actions"]:
                    sub_act = oldmap[a["file"]]
                    subdata = sub_act["data"]
                    if "tags" in subdata:
                        for t in subdata["tags"]:
                            tags_dict[t] = 1

                actdata["tags"] = tags_dict.keys()
            except KeyError:
                # Not all of the dependencies were loaded, I guess.
                pass
        """
        
        self.actmap = oldmap

    def _get_wsdl_data(self, act):
        """
        Reads cached WSDL information for the specified WSDL action
        entry, if any exists. If the information does not exist, or is
        not up to date, refreshes the cache entry. Sets "data" in
        "act".
        """
        wsdl_file = act["file"]

        cache_file = Config.wsdl_dir + Util.wsdl_cache_file(wsdl_file)
        data = {}
        is_current = False
        try:
            data = self._get_action_data(cache_file)
            is_current = (os.path.getmtime(cache_file) >
                          os.path.getmtime(Config.wsdl_dir + wsdl_file))
        except IOError:
            pass

        if not is_current:
            moddata = self.get_module("wsdl", act).get_data()

            ## Any info we got actually comes from the WSDL,
            ## so we'll have it overwrite any existing data.
            data.update(moddata)

            fp = open(cache_file, "w")
            try:
                fp.write("data = " + repr(data))
            finally:
                fp.close()

        ## Lets have some tags.
        if "tags" not in data:
            data["tags"] = ['web service', 'WSDL']

        return data

    def _import_action(self, act):
        modname = "s4all.actions." + act["file"][:-3]
        mod = Util.import_named(modname)
        act["module"] = mod

    def _get_action_data(self, file):
        """
        Tries to get just "data" from the specified file.
        """
        elocals = {}
        code = self._read_action_data(file)
        exec(code, {}, elocals)
        #print repr(elocals)
        return elocals["data"]

    def _read_action_data(self, file):
        """
        A poor man's solution for reading everything up to and including
        a statement that assigns to a variable called "data".
        For now, we just use a marker to indicate where to stop reading,
        and we read everything if it's not there.
        """
        fp = open(file)
        lines = []
        try:
            while True:
                line = fp.readline()
                if line == "": # EOF
                    break
                if line.startswith("#---"): # end "data" section
                    break
                lines.append(line)
        finally:
            fp.close()
        return "".join(lines)

    def __error_no_actions(self):
        appuifw.note(u"No local actions",
                     "error")

############################################################
#
#                     WSDLWrapper
#
############################################################

class WSDLWrapper:

    def __init__(self, proxy, file):
        
        self.SOAPProxy = proxy

        # In order to be able to use this class at the same way we are using
        # phones actions we want to put the name of the file here in which
        # this wsdl-file was loaded from.
        self.__file__ = file

    def get_data(self):
        data = { "type": "wsdl",
                 "parameters": [self._get_SOAP_methods()] }
        name = self._get_SOAP_name()
        doc = self._get_SOAP_doc()

        #input = self._get_SOAP_input()
        #output = self._get_SOAP_output()
        #Getting the input- and output types might be next-to-impossible to automate.

        #self.SOAPproxy.methods["DailyDilbertImage"].getOutParams() returns only the
        #complex type name, which is the output. The real output type need to be
        #fetched from for example
        #self.SOAPproxy.types[0].elements["DailyDilbertImageResponse"].content.content.content[0].type
        #where there might be multiple output parameters in the last content-list.

        input = "anything"
        output = "anything"
        if name is not None:
            data["name"] = unicode(name)
        if doc is not None:
            data["doc"] = unicode(doc)
        data["input"] = input
        data["output"] = output
        return data

    def run(self, input, p):

        selection = p["method"]

        call = getattr(self.SOAPProxy, selection)

        number_of_params = len(self.SOAPProxy.methods[selection].inparams)

        input[0] = call(input[0])
        return input
                
    def _get_SOAP_doc(self):
        """
        We want to have a string in the documentation
        """
        doctext = self.SOAPProxy.wsdl.services[0].documentation
        if doctext is None:
            return None
        if len(doctext) == 0:
            return None
        return doctext

    def _get_SOAP_name(self):
        if self.SOAPProxy.wsdl.services[0].name is not None:
            return unicode(self.SOAPProxy.wsdl.services[0].name)

    def _get_SOAP_methods(self):
        methods = self.SOAPProxy.methods.keys()
        method_tuples = [unicode(m) for m in methods]

        parameters = ('method', 'combo', (method_tuples, 0))
        return parameters

    def _get_SOAP_input(self):
	
	return "anything"

    def _get_SOAP_output(self):
	return "anything"
