# composer_workflow.py
#
# Copyright 2006-2007 Helsinki Institute for Information Technology
# and the authors.
#
# Authors: Ville Mäntysaari <ville.mantysaari@hiit.fi>
#          Pekka Kanerva <kanerva@hiit.fi>
#          Tero Hasu <tero.hasu@hut.fi>
#

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import e32
import appuifw

from s4all.composer.util import Str, Config, Util
from s4all.composer.runtime import *



class ActionFailure(Exception):
    # excinfo:: As returned by sys.exc_info().
    def __init__(self, act, excinfo):
        self.act = act
        self.exc = excinfo

    def __str__(self):
        return repr(self.value)



############################################################
#
#                     Workflow
#
############################################################


class Workflow:
    """
    A class encapsulating a single workflow
    """
    
    def __init__(self, name, gui, a_store):

        self.name = name
        self.gui = gui
        self.actions = []
        self.modified = False
        self.input = {}
        self.a_store = a_store

        if self.gui is not None:
            self.wflow_menu = [(u"Insert action", self.insert_action_handler),
                               (u"Append action", self.append_action_handler),
                               (u"Remove action", self.remove_action_handler),
                               (u"Configure", self.configure_action_handler),
                               (u"Show information", self.show_action_doc_handler),
                               (u"Run workflow", self.run_handler),
                               (u"Save workflow", self.save_wflow_handler),
                               (u"Save workflow as...", self.save_as_handler),
                               (u"Search registry", self.search_actions_handler),
                               (u"Close workflow", self.close_wflow_handler)]


    def close_wflow_handler(self):
        if self.modified:
            if appuifw.query(u'Workflow modified, save?', 'query'):
                self.save_wflow_handler()
        self.gui.close_wflow_handler()

    def load(self, basename):
        """
        Any existing data will be replaced with the loaded data.
        """
        action = self.a_store.get_action_by_basename(basename)
        loaded = action["data"]["actions"]

        action_list = None
        try:
            action_list = [
                {"action": self.a_store.get_action_by_basename(a["file"]),
                 "paramvals": a["paramvals"]}
                for a in loaded ]
        except KeyError:
            # XXX note that the above will not detect if child composites
            # do not have all their constituent actions loaded
            appuifw.note(u"Workflow contains an action which "
                         u"is not stored locally", "error")
            raise

        self.actions = action_list
        self.modified = False

    def append_action_handler(self):
        new_action = self.a_store.interactive_find_action()
        if new_action is None:
            return
        self._add_action(new_action)

    def insert_action_handler(self):
        new_action = self.a_store.interactive_find_action()
        if new_action is None:
            return
        insert_index = self.gui.main_view.current()
        self._add_action(new_action, insert_index)

    def remove_action_handler(self):
        if self.is_empty():
            self.__error_no_actions()
            return

        index = self.gui.main_view.current()
        del self.actions[index]

        # What about the input/output types of the actions that are neighbours
        # to this removed one?

        self.modified = True
            
        self.gui.main_view.set_list(self.get_action_names())

    def back_to_wflow_handler(self):
        appuifw.app.title = self.name
        appuifw.app.body = self.gui.main_view
        appuifw.app.menu = self.wflow_menu

    def configure_action_handler(self):
        if self.is_empty():
            self.__error_no_actions()
            return

        index = self.gui.main_view.current()
        action_entry = self.actions[index]
        form_fields = self._current_config(action_entry)

        if len(form_fields) == 0:
            appuifw.note(u"No configurable parameters available", "error")
            return

        old_config = [ entry[2] for entry in form_fields ]
        config_form = appuifw.Form(form_fields,
                                   appuifw.FFormEditModeOnly|
                                   appuifw.FFormDoubleSpaced)
        # This method seems to always return None.
        config_form.execute()

        # if 'type' field of the 3-tuple Form is returning happens to
        # be of 'combo', the value is a two tuple of form ( [list],
        # index ). Now, for some reason the index field of the two
        # tuple is returned as long (as 'nnnL'), which causes the Form
        # to crash when a saved workflow configuration is read from a
        # file... The same goes for "number" field types.

        new_config = []
        for n,t,v in config_form:
            if type(v) == tuple:
                new_config.append((v[0], int(v[1])))
            elif type(v) == long:
                new_config.append(int(v))
            else:
                new_config.append(v)

        if (new_config != old_config):
            self.modified = True
        action_entry["paramvals"] = new_config

    def _current_config(self, entry):
        par_inf = entry["action"]["data"].get("parameters", [])
        par_val = entry["paramvals"]
        return [(unicode(x[0]), x[1], y)
                for x, y
                in zip(par_inf, par_val)]

    def _config_list_to_map(self, list):
        map = {}
        for n,t,v in list:
            if t == "combo":
                v = v[0][v[1]]
            map[str(n)] = v
        return map

    def __error_no_actions(self):
        appuifw.note(u"Workflow doesn't contain any actions",
                     "error")

    def __import_whole_action(self, action):
        acttype = action["data"]["type"]

        if acttype in ["python", "wsdl", "complex"]:
            self.a_store.get_module(acttype, action)
        elif acttype == "wflow":
            actlist = action["data"]["actions"]
            filelist = [ subact["file"] for subact in actlist ]
            for basename in filelist:
                subact = self.a_store.get_action_by_basename(basename)
                self.__import_whole_action(subact)

    def __run_whole_action(self, entry, input):
        action = entry["action"]
        acttype = action["data"]["type"]

        if acttype in ["python", "wsdl"]:

            try:
                params = self._current_config(entry)
                paramtype = action["data"].get("paramtype", "dict")
                argstype = action["data"].get("argstype", "tuples")
                if paramtype == "dict":
                    params = self._config_list_to_map(params)
                if argstype == 'dict':
                    input = action["module"].run(input, params)
                else:
                    input["input"] = action["module"].run(input["input"], params)
                return input
            except StopFlow, e:
                raise # propagate up
            except Exception, e:
                raise ActionFailure(action, sys.exc_info())
        elif acttype == "wflow":
            actlist = action["data"]["actions"]
            for subdata in actlist:
                basename = subdata["file"]
                subact = self.a_store.get_action_by_basename(basename)
                subentry = {"action": subact,
                            "paramvals": subdata["paramvals"]}
                input = self.__run_whole_action(subentry, input)
            return input
        elif acttype == "control":
            params = self._current_config(entry)
            actions_to_loop = int(params[0][2])
            actlist = self.actions[self.actions.index(entry)+1:self.actions.index(entry)+actions_to_loop]

            input = [input, actlist]

            try:
                params = self._current_config(entry)
                paramtype = action["data"].get("paramtype", "dict")
                argstype = action["data"].get("argstype", "dict")
                if argstype == "dict":
                    params = self._config_list_to_map(params)
                output = action["module"].run(self, input, params)
            except StopFlow, e:
                raise # propagate up
            except Exception, e:
                raise ActionFailure(action, sys.exc_info())

	    raise EndLoop(output) # ..

    def __check_cycles(self, action, checked):
        acttype = action["data"]["type"]
        if acttype == "wflow":
            if action in checked:
                appuifw.note(u'Cycle in workflow!', 'error')
                raise ValueError
            else:
                checked.append(action)
            actlist = action["data"]["actions"]
            filelist = [ subact["file"] for subact in actlist ]
            for basename in filelist:
                subact = self.a_store.get_action_by_basename(basename)
                self.__check_cycles(subact, checked)

    def run_handler(self):
        # Ensure that all the code to execute has been loaded.
        # In case of composite actions, we may need to import
        # multiple modules (recursively).

        checked = []

        for act in self.actions:
            try:
                self.__check_cycles(act["action"], checked)
            except ValueError:
                return

        for entry in self.actions:
            self.__import_whole_action(entry["action"])

        # In order to save more things for actions somewhere in the
        # future we are going to use a dictionary as the input list

        self.input["input"] = [""]
        
        old_name = ""
        try:
            for entry in self.actions:
                index = self.actions.index(entry)
                old_name = entry["action"]["data"]["name"]
                entry["action"]["data"]["name"] = u'-> ' + entry["action"]["data"]["name"]
                self.actions[index] = entry
                self.gui.main_view.set_list(self.get_action_names())
                self.input = self.__run_whole_action(entry, self.input)
                entry["action"]["data"]["name"] = old_name
                self.gui.main_view.set_list(self.get_action_names())
        except StopFlow, e:
            e.act["data"]["name"] = old_name
            self.input["input"] = e.value
        except ActionFailure, e:
            e.act["data"]["name"] = old_name
            name = e.act["data"]["name"]
            errtext = "Error in action %s" % name
            appuifw.note(unicode(errtext), "error")

            #type, value, trace = sys.exc_info()
            import traceback
            print errtext
            traceback.print_exception(*e.exc)

            return

        except EndLoop, e:
            self.input["input"][0] = e.value[0]
            continue_from = e.value[1]

            rest_of_actions = self.actions[continue_from+1:]

            try:
                for entry in rest_of_actions:
                    self.input = self.__run_whole_action(entry, self.input)
            except StopFlow, e:
                self.input["input"] = e.value
            except ActionFailure, e:
                name = e.act["data"]["name"]
                errtext = "Error in action %s" % name
                appuifw.note(unicode(errtext), "error")

                import traceback
                print errtext
                traceback.print_exception(*e.exc)

                return

        appuifw.note(unicode(repr(self.input["input"][0])), "info")

    def show_action_doc_handler(self):
        if self.is_empty():
            self.__error_no_actions()
            return

        index = self.gui.main_view.current()
        data = self.actions[index]["action"]["data"]
        if "doc" not in data:
            appuifw.note(u"Action is undocumented",
                         "error")
            return
        doctext = data["doc"]

        action_text = appuifw.Text()
        action_text.set(doctext)
        action_text.set_pos(0)
        appuifw.app.body = action_text
        appuifw.app.title = u"Action doc"
        appuifw.app.menu = [(u"Back", self.back_to_wflow_handler)]

    def save_as_handler(self):

        if self.is_empty():
            self.__error_no_actions()
            return

        old_name = self.name
        new_name = appuifw.query(u'New workflow name',
                                 'text',
                                 self.name)
        if new_name is None:
            return
        basename = Util.wflow_basename_from_name(new_name)
        if not Util.wflow_exists_check(basename):
            appuifw.note(u"Workflow not saved", "info")
            return

        self.name = new_name
        appuifw.app.title = self.name

        if old_name != new_name:
            self.modified = True

        self.save_wflow_handler()

    def save_wflow_handler(self):
        savemodel = {"type": "wflow",
                     "name": self.name,
                     "doc": u"A composite action.",
                     "actions": [ {"file": a["action"]["file"],
                                   "paramvals": a["paramvals"]}
                                  for a in self.actions ] }
        
        tags_dict = {}
        for a in self.actions:
            subdata = a["action"]["data"]
            if "tags" in subdata:
                for t in subdata["tags"]:
                    tags_dict[t] = 1
        savemodel["tags"] = tags_dict.keys()

        savemodel["input"] = self.actions[0]["action"]["data"]["input"]
        savemodel["output"] = self.actions[len(self.actions)-1]["action"]["data"]["output"]

        basename = Util.wflow_basename_from_name(self.name)
        self.a_store.store_action(basename, savemodel)

        wflow_file = open(Config.actions_dir + basename, "w")
        try:
            wflow_file.write('data = ')
            wflow_file.write(repr(savemodel))
            wflow_file.write("\n")
        finally:
            wflow_file.close()

        self.modified = False
        appuifw.note(u"Saved workflow", "info")

    def search_actions_handler(self):
        self.a_store._search_actions_handler()

    def get_action_names(self):
        if self.is_empty():
            return [ Str.no_actions ]
        return [ a["action"]["data"]["name"] for a in self.actions ]

    def is_empty(self):
        return (len(self.actions) == 0)


    #
    # Private functions
    #

    def _add_action(self, new_action, insert_index = None):
        entry = self.__entry_for_action(new_action)

        if insert_index is None:
            self.actions.append(entry)
            ind = self.actions.index(entry)
            if ind is not 0:
                action_before_this = self.actions[ind-1]
                if (entry["action"]["data"]["input"] == action_before_this["action"]["data"]["output"]) or (entry["action"]["data"]["input"] == 'anything'):
                    print ""
                else:
                    self.actions.remove(entry)
                    appuifw.note(u"Input and output don't match!", 'error')
        else:
            self.actions.insert(insert_index, entry)
            if insert_index is not 0:
                action_before_this = self.actions[insert_index-1]
                if entry["action"]["data"]["input"] == action_before_this["action"]["data"]["output"]:
                    print ""
                else:
                    self.actions.remove(entry)
                    appuifw.note(u"Input and output don't match!", 'error')

        self.modified = True

        self.gui.main_view.set_list(self.get_action_names())

    def __entry_for_action(self, action):
        par_val = [v for n,t,v in action["data"].get("parameters", [])]
        entry = {"action": action,
                 "paramvals": par_val}
        return entry

